from typing import Any, Callable, Final, Generic, Tuple, TypeVar
from typing_extensions import Protocol, Self
from .lexer import BaseLexer, Token


class SupportGetCommand(Protocol):
    def __getitem__(self, __key: str) -> Callable: ...


T_CmdSet = TypeVar("T_CmdSet", bound=SupportGetCommand)
T_Lexer = TypeVar("T_Lexer", bound=BaseLexer)


class Parser(Generic[T_Lexer, T_CmdSet]):
    lexer: Final[T_Lexer]
    command_set: Final[T_CmdSet]

    def __init__(self, lexer: T_Lexer, command_set: T_CmdSet) -> None: ...
    def push(self, token: Token) -> None: ...
    def pop(self) -> Token: ...
    def parse_args(self) -> Tuple[tuple, dict]: ...
    def exec_once(self) -> Any: ...
    def exec(self) -> None: ...
    def eof(self) -> bool: ...
    def __iter__(self) -> Self: ...
    def __next__(self) -> Any: ...


"""
Copyright 2022 Ovizro

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from io import StringIO
import os
import sys
from argparse import ArgumentParser
from traceback import print_exc
from typing import Any, Callable, Optional

from .lexer import BaseLexer, FileLexer, StringLexer
from .klvm import KoiLang, KoiLangMeta, kola_command, kola_env, kola_text
from .exception import KoiLangCommandError, KoiLangError

from . import __version__


def _load_script(path: str, encoding: str = "utf-8") -> KoiLangMeta:
    vdict = {}
    with open(path, encoding=encoding) as f:
        exec(
            compile(f.read(), path, "exec"),
            vdict
        )
    for i in vdict.values():
        if isinstance(i, KoiLangMeta) and i is not KoiLang:
            return i
    else:
        raise TypeError("no KoiLang command set found")


class _CommandDebugger(KoiLang):
    def __getitem__(self, key: str) -> Callable[..., None]:
        def wrapper(*args, **kwds) -> None:
            print(f"cmd: {key} with args {args} kwds {kwds}")
        return wrapper


class KoiLangMain(KoiLang):
    def __init__(self, parent: Optional[KoiLang] = None) -> None:
        super().__init__(parent)
        self.vars = {}

    @kola_command
    def version(self) -> None:
        print(__version__)
    
    @kola_command
    def license(self) -> None:
        print(__doc__)
    
    @kola_command
    def raises(self) -> None:
        raise KoiLangCommandError
    
    @kola_command
    def echo(self, *text: str) -> None:
        print(' '.join(text))
    
    @kola_command("get")
    def get_(self, key: str) -> None:
        print(self.get_var(key))
    
    @kola_command
    def set(self, **kwds) -> None:
        self.vars.update(kwds)
    
    @kola_command
    def mkdir(self, dir: str, mode: int = 777) -> None:
        os.mkdir(dir, mode)
    
    @kola_env
    def open(self, path: str, mode: str = "w", *, encoding: str = "utf-8") -> None:
        if hasattr(self, "file"):
            self.file.close()
        self.file = open(path, mode, encoding=encoding)
    
    @open.exit_command
    def close(self) -> None:
        self.file.close()
        del self.file
    
    @kola_command
    def remove(self, path: str) -> None:
        os.remove(path)
    
    @kola_command
    def load(self, path: str, type: str = "kola", *, encoding: str = "utf-8") -> None:
        if type == "kola":
            self.parse_file(path)
        elif type == "script":
            cmd_set = _load_script(path, encoding=encoding)
            self.push(cmd_set.__name__, cmd_set())
        else:
            raise ValueError("load type only supports 'kola' and 'script'")

    @kola_command
    def reset(self) -> None:
        while self._stack.name != "__init__":
            self.pop()
            
    @kola_text
    def text(self, text: str) -> None:
        s = StringIO()
        i = 0
        while i < len(text):
            if text[i] == '$':
                i += 1
                j = i
                while text[i] not in ['\n', ' ', '\t']:
                    i += 1
                    if i >= len(text):
                        break
                key = text[j: i]
                val = self.get_var(key) or ''
                s.write(str(val))
            else:
                s.write(text[i])
                i += 1
        text = s.getvalue()
        if hasattr(self, "file"):
            self.file.write(text)
        else:
            print(f"::{text}")
    
    @kola_command
    def exit(self, code: int = 0) -> None:
        sys.exit(code)
    
    def get_var(self, key: str) -> Any:
        if key == "__name__":
            return self.__class__.__name__
        elif key == '__top__':
            return self.top[0]
        elif key == "__dir__":
            return ', '.join(i.__name__ for i in self.__class__.__command_field__)
        else:
            return self.vars.get(key, None)


def _read_stdin() -> str:
    sys.stdout.write("$kola: ")
    sys.stdout.flush()
    s = sys.stdin.readline()
    while s.endswith("\\\n"):
        sys.stdout.write("$...: ")
        sys.stdout.flush()
        s += sys.stdin.readline()
    return s


if __name__ == "__main__":
    parser = ArgumentParser("kola")
    parser.add_argument("file", default=None, nargs="?")
    parser.add_argument("-i", "--inline", help="parse inline string")
    parser.add_argument("-s", "--script", help="parser script")
    parser.add_argument("-d", "--debug", help="dubugger type", choices=["token", "command"])
    parser.add_argument("--encoding", help="file encoding", default="utf-8")

    namespace = parser.parse_args()

    if namespace.file:
        lexer = FileLexer(namespace.file)
    elif namespace.inline:
        lexer = StringLexer(namespace.inline)
    else:
        lexer = None

    runner_type = "Runner"
    if namespace.debug == "token":
        runner_type = "Token Debugger"
        if lexer is None:
            print(f"KoiLang {runner_type} {__version__} on Python {sys.version}")
            lexer = BaseLexer()
        while True:
            try:
                for i in lexer:
                    print(i)
                break
            except KeyboardInterrupt:
                break
            except KoiLangError:
                print_exc()
    else:
        if namespace.debug == "command":
            command_cls = _CommandDebugger
            runner_type = "Command Debugger"
        elif namespace.script:
            command_cls = _load_script(namespace.script)
        else:
            command_cls = KoiLangMain

        command_set = command_cls()
        if lexer:
            command_set.parse(lexer)
        else:
            print(f"KoiLang {runner_type} {__version__} on Python {sys.version}")
            while True:
                try:
                    command_set.parse(_read_stdin())
                except KeyboardInterrupt:
                    break
                except KoiLangError:
                    print_exc()



from libc.stdio cimport FILE
from libc.stdint cimport uint8_t, uint64_t
from cpython cimport PyObject


cdef extern from "<stdarg.h>":
    ctypedef struct va_list
    void va_start(va_list ap, const char* last_arg) nogil
    void va_end(va_list ap) nogil

cdef extern from "Python.h":
    PyObject* _PyType_Lookup(type t, str name)
    
    ctypedef uint64_t Py_UCS4
    
    str PyUnicode_FromFormat(const char* format, ...)
    Py_UCS4 PyUnicode_READ_CHAR(str unicode, Py_ssize_t index)
    int PyUnicode_WriteChar(
        str unicode, Py_ssize_t index, Py_UCS4 character) except -1

    enum PyUnicode_Kind:
        PyUnicode_WCHAR_KIND = 0
        PyUnicode_1BYTE_KIND = 1
        PyUnicode_2BYTE_KIND = 2
        PyUnicode_4BYTE_KIND = 4

    ctypedef struct _PyUnicodeWriter:
        PyObject* buffer
        void* data
        PyUnicode_Kind kind
        Py_UCS4 maxchar
        Py_ssize_t size
        Py_ssize_t pos

        # minimum number of allocated characters (default: 0)
        Py_ssize_t min_length

        # minimum character (default: 127, ASCII)
        Py_UCS4 min_char

        # If non-zero, overallocate the buffer (default: 0).
        unsigned char overallocate

        # If readonly is 1, buffer is a shared string (cannot be modified)
        # and size is set to 0.
        unsigned char readonly_ "readonly"
    
    void _PyUnicodeWriter_Init(_PyUnicodeWriter *writer)
    int _PyUnicodeWriter_Prepare(_PyUnicodeWriter *writer, Py_ssize_t length, Py_UCS4 maxchar) except -1
    int _PyUnicodeWriter_PrepareKind(_PyUnicodeWriter *writer, PyUnicode_Kind kind) except -1
    int _PyUnicodeWriter_WriteChar(_PyUnicodeWriter *writer, Py_UCS4 ch) except -1
    int _PyUnicodeWriter_WriteStr(_PyUnicodeWriter *writer, str string) except -1
    int _PyUnicodeWriter_WriteASCIIString(_PyUnicodeWriter *writer, const char* string, Py_ssize_t len) except -1
    str _PyUnicodeWriter_Finish(_PyUnicodeWriter *writer)
    void _PyUnicodeWriter_Dealloc(_PyUnicodeWriter *writer)


cdef extern from "_cutil.h":
    enum TokenSyn:
        CMD
        CMD_N
        TEXT
        LITERAL
        STRING
        NUM
        NUM_H
        NUM_B
        NUM_F
        CLN
        CMA
        SLP
        SRP
        ANNOTATION
    const uint8_t yy_goto[7][8]

    void kola_set_error(object exc_type, int errorno, const char* filename, int lineno, const char* text) except *
    void kola_set_errcause(object exc_type, int errorno, const char* filename, int lineno, const char* text, object cause) except *

    FILE* kola_open(object raw_path, PyObject** out, const char* mod) except NULL

    const char* get_type_name(object obj) nogil
    const char* get_type_qualname(object obj) nogil
    const char* unicode2string(str __s, Py_ssize_t* s_len) except NULL
    str decode_escapes(const char* string, Py_ssize_t len)
    PyObject* filter_text(str string) except NULL



import os
from typing import Any, Dict, Final, List, Union
from typing_extensions import Protocol, runtime_checkable, Self


WF_BASE_ITEM: int
WF_COMPLEX_ITEM: int
WF_ARG_ITEM: int
WF_FULL_CMD: int


@runtime_checkable
class WriterItemLike(Protocol):
    def __kola_write__(self, __writer: BaseWriter, __level: int) -> None: ...


class BaseWriterItem(object):
    def __kola_write__(self, writer: BaseWriter, level: int) -> None: ...


class FormatItem(BaseWriterItem):
    value: Final[Any]
    spec: Final[str]

    def __init__(self, value: Any, spec: str) -> None: ...


class ComplexArg(BaseWriterItem):
    name: Final[str]
    value: Final[Any]

    def __init__(self, name: str, value: Any, *, split_line: bool = ...) -> None: ...


WI_NEWLINE: BaseWriterItem


_BaseItem = Union[str, bytes, int, float, WriterItemLike]
_ComplexItem = Union[_BaseItem, List[_BaseItem], Dict[str, _BaseItem]]


class BaseWriter(object):
    indent: Final[int]

    def __init__(self, indent: int = 4, command_threshold: int = 1) -> None: ...
    def raw_write(self, text: str) -> None: ...
    def close(self) -> None: ...
    def inc_indent(self) -> None: ...
    def dec_indent(self) -> None: ...
    def newline(self) -> None: ...
    def write_text(self, text: str) -> None: ...
    def write_command(self, __name: Union[str, int], *args: _BaseItem, **kwds: _ComplexItem) -> None: ...
    def write(self, command: Union[str, WriterItemLike]) -> None: ...
    @property
    def closed(self) -> bool: ...
    def __enter__(self) -> Self: ...
    def __exit__(self, exc_type, exc_inc, traceback) -> None: ...


class FileWriter(BaseWriter):
    path: Final[Union[str, bytes, os.PathLike]]
    encoding: Final[str]

    def __init__(self, __path: Union[str, bytes, os.PathLike], encoding: str = "utf-8",
                 indent: int = ..., command_threshold: int = ...) -> None: ...


class StringWriter(BaseWriter):
    def getvalue(self) -> str: ...



class KoiLangError(Exception):
    """
    Base exception for marktext compiler
    """
    __slots__ = []


class KoiLangSyntaxError(KoiLangError):
    """
    Raised when failing to parse text
    """
    __slots__ = []


class KoiLangCommandError(KoiLangError):
    """
    Raised when errors occure during executing commands
    """
    __slots__ = []



from libc.stdio cimport FILE
from libc.stdint cimport uint8_t
from ._cutil cimport *


cdef enum ItemLevel:
    BASE_ITEM
    COMPLEX_ITEM
    ARG_ITEM
    FULL_CMD


cdef class BaseWriterItem(object):
    cpdef void __kola_write__(self, BaseWriter writer, ItemLevel level) except *


cdef class FormatItem(BaseWriterItem):
    cdef readonly:
        object value
        str spec
    cpdef void __kola_write__(self, BaseWriter writer, ItemLevel level) except *


cdef class ComplexArg(BaseWriterItem):
    cdef readonly:
        str name
        object value
        bint split_line
    cpdef void __kola_write__(self, BaseWriter writer, ItemLevel level) except *


cdef class NewlineItem(BaseWriterItem):
    cpdef void __kola_write__(self, BaseWriter writer, ItemLevel level) except *


cdef NewlineItem i_newline


cdef class BaseWriter:
    cdef Py_ssize_t cur_indent
    cdef readonly:
        uint8_t indent
        int command_threshold
    cdef public bint line_beginning

    cpdef void raw_write(self, str text) except *
    cdef void raw_write_string(self, const char* string, Py_ssize_t length = *) except *
    cdef void raw_write_char(self, char ch) except *
    cpdef void close(self)
    cpdef void inc_indent(self)
    cpdef void dec_indent(self) except *
    cdef void _write_indent(self) except *
    cdef void _write_prefix(self, Py_ssize_t length) except *
    cpdef void newline(self, bint concat_prev = *) except *
    cpdef void prepare(self) except *
    cdef void _write_text(self, str text) except *


cdef class FileWriter(BaseWriter):
    cdef FILE* fp
    cdef readonly:
        object path
        str encoding
    cpdef void close(self)
    cpdef void prepare(self) except *
    cpdef void raw_write(self, str text) except *
    cdef void raw_write_string(self, const char* string, Py_ssize_t length = *) except *
    cdef void raw_write_char(self, char ch) except *


cdef class StringWriter(BaseWriter):
    cdef:
        bint _closed
        _PyUnicodeWriter writer
    
    cpdef void close(self)
    cpdef void prepare(self) except *
    cpdef void raw_write(self, str text) except *
    cdef void raw_write_string(self, const char* string, Py_ssize_t length = *) except *
    cdef void raw_write_char(self, char ch) except *
    cpdef str getvalue(self)



import os
from typing import Any, Final, Union, final
from typing_extensions import Self


S_CMD: int
S_CMD_N: int
S_TEXT: int
S_LITERAL: int
S_STRING: int
S_NUM: int
S_NUM_H: int
S_NUM_B: int
S_NUM_F: int
S_CLN: int
S_CMA: int
S_SLP: int
S_SRP: int
S_ANNOTATION: int


@final
class Token:
    """
    Token used in lexer

    Don't instantiate this class directly unless you make
    sure enough arguments provided.
    """

    syn: Final[int]
    val: Final[Any]

    def __new__(cls, syn: int, val: Any = ..., *, lineno: int = ..., raw_val: bytes = ...) -> Self: ...
    def get_flag(self) -> int: ...


class BaseLexer:
    lineno: Final[int]
    stat: Final[int]

    def __init__(self, *, encoding: str = ..., stat: int = 0, command_threshold: int = 1) -> None: ...
    def close(self) -> None: ...
    @property
    def filename(self) -> str: ...
    @property
    def closed(self) -> bool: ...
    def __iter__(self) -> Self: ...
    def __next__(self) -> Token: ...


class FileLexer(BaseLexer):
    def __init__(self, __path: Union[str, bytes, os.PathLike], *, encoding: str = ...,
                 stat: int = 0, command_threshold: int = 1) -> None: ...
    @property
    def filename(self) -> Union[str, bytes, os.PathLike]: ...


class StringLexer(BaseLexer):
    content: Final[bytes]

    def __init__(self, content: Union[str, bytes], *, encoding: str = ...,
                 stat: int = 0, command_threshold: int = 1) -> None: ...



from libc.stdio cimport stdin, FILE, fopen, fclose, EOF
from cpython cimport PyObject

from ._cutil cimport *


cdef extern from *:
    struct yy_buffer_state:
        pass
    ctypedef yy_buffer_state *YY_BUFFER_STATE

    char* yytext
    int yyleng
    int yylineno
    YY_BUFFER_STATE yy_current_buffer

    int get_stat()
    void set_stat(int stat)

    int yylex(int command_threshold)
    void yyrestart(FILE *input_file) nogil
    YY_BUFFER_STATE yy_create_buffer(FILE* file, int size) nogil
    YY_BUFFER_STATE yy_scan_buffer(char * text, Py_ssize_t size) nogil
    YY_BUFFER_STATE yy_scan_bytes(const char * text, int len) nogil
    void yy_switch_to_buffer(YY_BUFFER_STATE new_buffer) nogil
    void yy_flush_buffer(YY_BUFFER_STATE b) nogil
    void yy_delete_buffer(YY_BUFFER_STATE b) nogil


cdef class Token:
    cdef:
        Token next     # used in grammar parser
    cdef readonly:
        TokenSyn syn
        object val
        bytes raw_val
        int lineno
    
    cpdef int get_flag(self)
    

cdef class BaseLexer:
    cdef:
        const char* _filename
        YY_BUFFER_STATE buffer
    cdef readonly:
        str encoding
        int lineno
        int stat
        int command_threshold

    cpdef void close(self)
    cdef void set_error(self) except *
    cdef void ensure(self)
    cdef (int, const char*, Py_ssize_t) next_syn(self)
    cdef Token next_token(self)


cdef class FileLexer(BaseLexer):
    cdef:
        object _filenameo
        bytes _filenameb
        FILE* fp
    
    cpdef void close(self)


cdef class StringLexer(BaseLexer):
    cdef readonly bytes content


from ._cutil cimport *
from .lexer cimport Token, BaseLexer


cdef extern from *:
    str PyUnicode_FromFormat(const char*, ...)


cdef class Parser:
    cdef:
        Token t_cache
        Token stack_top
        uint8_t stat
    cdef readonly:
        BaseLexer lexer
        object command_set

    cpdef void push(self, Token n)
    cpdef Token pop(self)
    cdef void set_error(self, int errorno = *, bint recovery = *) except *
    cpdef tuple parse_args(self)
    cpdef object exec_once(self)
    cpdef void exec(self) except *


from functools import partial
import os
from types import MethodType
from typing import (Any, Callable, Dict, Generator, Iterable, Mapping,
                    NamedTuple, Optional, Set, Tuple, Type, Union, cast,
                    overload)
from typing_extensions import Literal, Self

from .exception import KoiLangCommandError
from .lexer import BaseLexer, FileLexer, StringLexer
from .parser import Parser
from .writer import BaseWriter, FileWriter, StringWriter


class EnvNode(NamedTuple):
    name: str
    command_set: Union[Dict[str, Callable], "KoiLang"]
    next: Optional["EnvNode"] = None


class Command(object):
    __slots__ = ["__name__", "__func__", "method_type", "envs", "suppression", "alias", "writer_func"]

    def __init__(
        self,
        name: str,
        func: Callable,
        *,
        method: Literal["static", "class", "default"] = "default",
        envs: Union[Iterable[str], str] = tuple(),
        alias: Union[Iterable[str], str] = tuple(),
        suppression: bool = False
    ) -> None:
        self.__name__ = name
        self.__func__ = func
        self.method_type = method
        if envs and method != "default":
            raise KoiLangCommandError(
                "cannot accesss the environment for class method and static method"
            )
        self.envs = (envs,) if isinstance(envs, str) else tuple(envs)
        self.alias = (alias,) if isinstance(alias, str) else tuple(alias)
        self.suppression = suppression
        if self.__name__ == "@text":
            self.writer_func = BaseWriter.write_text
        else:
            self.writer_func = lambda writer, *args, **kwds: \
                writer.write_command(name, *args, **kwds)

    def bind(self, ins: Optional["KoiLang"], owner: Optional[Type["KoiLang"]] = None) -> Callable:
        """binding command function with an instance"""
        if self.method_type == "default" and ins is not None:
            return MethodType(self, ins)
        elif self.method_type == "class":
            return MethodType(self, owner or type(ins))
        else:
            return self
    
    def generate_commands(self, vmobj: "KoiLang", force: bool = False) -> Generator[Tuple[str, Callable], None, None]:
        if not self.suppression or force:
            bound_func = self.bind(vmobj)
            yield self.__name__, bound_func
            for i in self.alias:
                yield i, bound_func
    
    def writer(self, func: Callable) -> Self:
        self.writer_func = func
        return self

    def __get__(self, instance: Any, owner: type) -> Callable:
        return self.bind(instance, cast(Type[KoiLang], owner))

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        if self.method_type == "default" and self.envs:
            vmobj: KoiLang = args[0]
            env_name, _ = vmobj.top
            if env_name not in self.envs:
                raise KoiLangCommandError(f"unmatched environment {env_name}")
        return self.__func__(*args, **kwds)
    
    def __repr__(self) -> str:
        return f"<kola command {self.__name__} with {self.__func__}>"


class BaseEnv(Command):
    __slots__ = ["env_name"]

    def __init__(
        self,
        name: str,
        func: Callable,
        *,
        method: Literal["static", "class", "default"] = "default",
        env_name: Optional[str] = None,
        **kwds
    ) -> None:
        super().__init__(name, func, method=method, **kwds)
        self.env_name = env_name or name

    def bind(self, ins: "KoiLang", owner: Optional[Type["KoiLang"]] = None) -> Callable:
        if ins is not None:
            return MethodType(self, ins)
        return self
    
    def __call__(self, vmobj: "KoiLang", *args, **kwds) -> Any:
        if self.method_type == "static":
            return super().__call__(*args, **kwds)
        elif self.method_type == "class":
            return super().__call__(type(vmobj), *args, **kwds)
        else:
            return super().__call__(vmobj, *args, **kwds)

    def __repr__(self) -> str:
        return f"<kola command {self.__name__} with {self.__func__} on '{self.env_name}'>"


class EnvEnter(BaseEnv):
    __slots__ = ["command_set", "command_set_extra", "auto_pop", "__write_func"]

    def __init__(
        self,
        name: str,
        func: Callable,
        *,
        command_set: Union[Dict[str, Callable], "KoiLang", None] = None,
        auto_pop: bool = True,
        **kwds
    ) -> None:
        super().__init__(name, func, **kwds)
        self.command_set = command_set
        self.auto_pop = auto_pop
        self.command_set_extra: Set[Command] = set()
    
    def __call__(self, vmobj: "KoiLang", *args, **kwds) -> Any:
        if self.auto_pop and vmobj.top[0] == self.env_name:
            vmobj.pop()
        vmobj.push(self.env_name, self.command_set or {})
        for i in self.command_set_extra:
            vmobj.update(i.generate_commands(vmobj, True))
        return super().__call__(vmobj, *args, **kwds)

    @overload
    def exit_command(self, func: Callable[..., Any], **kwds) -> Command: ...
    @overload  # noqa: E301
    def exit_command(self, func: Optional[str] = None, *, method: Literal["static", "class", "default"] = ...,
                     alias: Union[Iterable[str], str] = ...) -> Callable[[Callable[..., Any]], Command]: ...
    def exit_command(  # noqa: E301
        self,
        func: Union[Callable, str, None] = None,
        **kwds
    ) -> Union[Callable[[Callable], Command], Command]:
        """
        Define a command to exit current environment
        """
        def wrapper(wrapped_func: Callable[..., Any]) -> Command:
            if isinstance(func, str):
                name = func
            else:
                name = wrapped_func.__name__
            cmd = EnvExit(name, wrapped_func, env_name=self.env_name, suppression=True, **kwds)
            self.command_set_extra.add(cmd)
            self.auto_pop = False
            return cmd
        if callable(func):
            return wrapper(func)
        else:
            return wrapper
    
    @overload
    def env_command(self, func: Callable[..., Any], **kwds) -> Command: ...
    @overload  # noqa: E301
    def env_command(
        self,
        func: Optional[str] = None,
        *,
        method: Literal["static", "class", "default"] = ...,
        envs: Union[Iterable[str], str] = ...,
        alias: Union[Iterable[str], str] = ...
    ) -> Callable[[Callable[..., Any]], Command]: ...
    def env_command(  # noqa: E301
        self,
        func: Union[Callable, str, None] = None,
        **kwds
    ) -> Union[Callable[[Callable], Command], Command]:
        """
        Define a command only can be used in current environment
        """
        def wrapper(wrapped_func: Callable[..., Any]) -> Command:
            if isinstance(func, str):
                name = func
            else:
                name = wrapped_func.__name__
            cmd = Command(name, wrapped_func, suppression=True, **kwds)
            self.command_set_extra.add(cmd)
            return cmd
        if callable(func):
            return wrapper(func)
        else:
            return wrapper
    
    @overload
    def env_text(self, func: Callable[..., Any], **kwds) -> Command: ...
    @overload  # noqa: E301
    def env_text(
        self,
        *,
        method: Literal["static", "class", "default"] = ...,
        envs: Union[Iterable[str], str] = ...,
        alias: Union[Iterable[str], str] = ...
    ) -> Callable[[Callable[..., Any]], Command]: ...
    def env_text(  # noqa: E301
        self,
        func: Optional[Callable] = None,
        **kwds
    ) -> Union[Callable[[Callable], Command], Command]:
        """
        Define a text command only can be used in current environment
        """
        def wrapper(wrapped_func: Callable[..., Any]) -> Command:
            cmd = Command("@text", wrapped_func, suppression=True, **kwds)
            self.command_set_extra.add(cmd)
            return cmd
        if callable(func):
            return wrapper(func)
        else:
            return wrapper
    
    @overload
    def env_number(self, func: Callable[..., Any], **kwds) -> Command: ...
    @overload  # noqa: E301
    def env_number(
        self,
        *,
        method: Literal["static", "class", "default"] = ...,
        envs: Union[Iterable[str], str] = ...,
        alias: Union[Iterable[str], str] = ...
    ) -> Callable[[Callable[..., Any]], Command]: ...
    def env_number(  # noqa: E301
        self,
        func: Optional[Callable] = None,
        **kwds
    ) -> Union[Callable[[Callable], Command], Command]:
        """
        Define a number command only can be used in current environment
        """
        def wrapper(wrapped_func: Callable[..., Any]) -> Command:
            cmd = Command("@number", wrapped_func, suppression=True, **kwds)
            self.command_set_extra.add(cmd)
            return cmd
        if callable(func):
            return wrapper(func)
        else:
            return wrapper
        
    @overload
    def env_annotation(self, func: Callable[..., Any], **kwds) -> Command: ...
    @overload  # noqa: E301
    def env_annotation(
        self,
        *,
        method: Literal["static", "class", "default"] = ...,
        envs: Union[Iterable[str], str] = ...,
        alias: Union[Iterable[str], str] = ...
    ) -> Callable[[Callable[..., Any]], Command]: ...
    def env_annotation(  # noqa: E301
        self,
        func: Optional[Callable] = None,
        **kwds
    ) -> Union[Callable[[Callable], Command], Command]:
        """
        Define a annotation command only can be used in current environment
        """
        def wrapper(wrapped_func: Callable[..., Any]) -> Command:
            cmd = Command("@annotation", wrapped_func, suppression=True, **kwds)
            self.command_set_extra.add(cmd)
            return cmd
        if callable(func):
            return wrapper(func)
        else:
            return wrapper


class EnvExit(BaseEnv):
    __slots__ = []

    def __init__(
        self,
        name: str,
        func: Callable,
        *,
        env_name: str,
        **kwds
    ) -> None:
        super().__init__(name, func, env_name=env_name, envs=env_name, **kwds)
    
    def __call__(self, vmobj: "KoiLang", *args: Any, **kwds: Any) -> Any:
        ret = super().__call__(vmobj, *args, **kwds)
        assert vmobj.pop()[0] == self.env_name
        return ret


class EnvClsEnter(BaseEnv):
    __slots__ = ["auto_pop", "command_set_class", "exit_entry"]

    def __init__(
        self,
        name: str,
        func: Callable,
        *,
        command_set_class: Type["KoiLang"],
        env_name: Optional[str] = None,
        exit_entry: Union[str, Iterable[str]] = tuple(),
        **kwds
    ) -> None:
        super().__init__(name, func, env_name=env_name or command_set_class.__name__, **kwds)
        self.auto_pop = not exit_entry
        self.command_set_class = command_set_class
        if isinstance(exit_entry, str):
            self.exit_entry = (exit_entry,)
        else:
            self.exit_entry = tuple(exit_entry)

    def __call__(self, vmobj: "KoiLang", *args, **kwds) -> Any:
        if self.auto_pop and vmobj.top[0] == self.env_name:
            vmobj.pop()

        cmd_set = self.command_set_class(parent=vmobj)
        vmobj.push(self.env_name, cmd_set)
        for i in self.exit_entry:
            func: Callable = getattr(cmd_set, i)
            exit_cmd = EnvExit(func.__name__, func, env_name=self.env_name, method="static")
            vmobj.update(exit_cmd.generate_commands(vmobj))
        return super().__call__(cmd_set, *args, **kwds)


class KoiLangBaseWriter(BaseWriter):
    def __init__(
        self,
        command_set: "KoiLangMeta",
        *args,
        indent: int = 4,
        **kwds
    ) -> None:
        self.__command_set__ = command_set
        super().__init__(
            *args,
            indent=indent,
            command_threshold=command_set.__command_threshold__,
            **kwds
        )
    
    def __getattr__(self, name: str) -> Any:
        attr = getattr(self.__command_set__, name)
        if isinstance(attr, Command):
            return partial(attr.writer_func, self)
        else:
            raise AttributeError(f"kola writer object has no attribute {name}")


class KoiLangFileWriter(KoiLangBaseWriter, FileWriter):
    __slots__ = ["__command_set__"]

    def __init__(
        self,
        command_set: "KoiLangMeta",
        *,
        path: Union[str, bytes, os.PathLike],
        indent: int = 4
    ) -> None:
        super().__init__(command_set, path, indent=indent,
                         encoding=command_set.encoding)


class KoiLangStringWriter(KoiLangBaseWriter, StringWriter):
    __slots__ = ["__command_set__"]


class KoiLangMeta(type):
    """
    Metaclass for KoiLang class
    """
    encoding: str
    __command_field__: Set[Command]
    __command_threshold__: int

    def __new__(cls, name: str, bases: Tuple[type, ...], attr: Dict[str, Any],
                command_threshold: int = 0, text_encoding: Optional[str] = None, **kwds: Any):
        __command_field__ = set()

        has_base = False
        for i in bases:
            if isinstance(i, KoiLangMeta):
                has_base = True
                __command_field__.update(i.__command_field__)
        for i in attr.values():
            if isinstance(i, Command):
                __command_field__.add(i)
        attr["__command_field__"] = __command_field__

        if command_threshold or not has_base:
            assert command_threshold >= 0
            attr["__command_threshold__"] = command_threshold or 1
        if text_encoding or not has_base:
            attr["encoding"] = text_encoding or "utf-8"

        return super().__new__(cls, name, bases, attr, **kwds)
    
    @staticmethod
    def eval_commands(field: Set[Command], ins: Any) -> Dict[str, Callable]:
        cmd_set = {}
        for i in field:
            cmd_set.update(i.generate_commands(ins))
        return cmd_set

    def get_command_set(self, instance: Any) -> Dict[str, Callable]:
        return self.eval_commands(self.__command_field__, instance)

    @overload
    def register_command(self, func: Callable[..., Any], **kwds) -> Command: ...
    @overload  # noqa: E301
    def register_command(
        self,
        func: Optional[str] = None,
        *,
        method: Literal["static", "class", "default"] = ...,
        envs: Union[Iterable[str], str] = ...,
        alias: Union[Iterable[str], str] = ...
    ) -> Callable[[Callable[..., Any]], Command]: ...
    def register_command(  # noqa: E301
        self,
        func: Union[Callable, str, None] = None,
        **kwds
    ) -> Union[Callable[[Callable], Command], Command]:
        def wrapper(wrapped_func: Callable[..., Any]) -> Command:
            if isinstance(func, str):
                name = func
            else:
                name = wrapped_func.__name__
            cmd = Command(name, wrapped_func, suppression=True, **kwds)
            self.__command_field__.add(cmd)
            return cmd
        if callable(func):
            return wrapper(func)
        else:
            return wrapper
    
    @overload
    def writer(self, *, indent: int = 4) -> KoiLangStringWriter: ...
    @overload
    def writer(self, path: Union[str, bytes, os.PathLike], *, indent: int = 4) -> KoiLangFileWriter: ...
    def writer(  # noqa" 301
        self,
        path: Union[str, bytes, os.PathLike, None] = None,
        *,
        indent: int = 4
    ) -> Union[KoiLangFileWriter, KoiLangStringWriter]:
        if path:
            return KoiLangFileWriter(self, path=path, indent=indent)
        else:
            return KoiLangStringWriter(self, indent=indent)
    
    def __repr__(self) -> str:
        return f"<kola command set '{self.__qualname__}'>"


class KoiLang(metaclass=KoiLangMeta):
    """
    Main class for KoiLang virtual machine.
    """
    __slots__ = ["_stack", "back"]

    back: Optional["KoiLang"]

    def __init__(self, parent: Optional["KoiLang"] = None) -> None:
        super().__init__()
        command_set = self.__class__.get_command_set(self)
        self._stack = EnvNode("__init__", command_set)
        self.back = parent
    
    def get_env(self, name: str) -> Union[Dict[str, Callable], "KoiLang"]:
        stack = self._stack
        while stack:
            if stack.name == name:
                return stack.command_set
            stack = stack.next
        raise KeyError(f"environment {name} not found")
    
    def get(self, key: str, default: Optional[Callable] = None) -> Optional[Callable]:
        return self.__get(key, default)
    
    def update(self, other: Union[Mapping[str, Callable], Iterable[Tuple[str, Callable]]], **kwds) -> None:
        self._stack.command_set.update(other, **kwds)
    
    def keys(self) -> Iterable[str]:
        stack = self._stack
        while stack:
            yield from stack.command_set.keys()
            stack = stack.next
    
    def values(self) -> Iterable[Callable]:
        stack = self._stack
        while stack:
            yield from stack.command_set.values()
            stack = stack.next
        
    def items(self) -> Iterable[Tuple[str, Callable]]:
        stack = self._stack
        while stack:
            yield from stack.command_set.items()
            stack = stack.next
    
    def push(self, name: str, commands: Union[Dict[str, Callable], "KoiLang", Set[Command]]) -> None:
        if isinstance(commands, set):
            commands = self.__class__.eval_commands(commands, self)
        elif isinstance(commands, KoiLang):
            commands.at_start()
        self._stack = EnvNode(name, commands, self._stack)
    
    def pop(self) -> Tuple[str, Union[Dict[str, Callable], "KoiLang"]]:
        if self._stack.next is None:
            raise KoiLangCommandError("cannot pop inital stack")
        top = self.top
        if isinstance(top[1], KoiLang):
            top[1].at_end()
        self._stack = self._stack.next
        return top
    
    def at_start(self, **kwds) -> None:
        """
        Parser initalize method. Called before parsing start.
        """
    
    def at_end(self, **kwds) -> None:
        """
        Parser finalize method. Called after parsing end.
        """

    def parse(self, lexer: Union[BaseLexer, str], **kwds: Any) -> None:
        """
        Parse kola text or lexer from other method.
        """
        if isinstance(lexer, str):
            lexer = StringLexer(
                lexer,
                encoding=self.__class__.encoding,
                command_threshold=self.__class__.__command_threshold__
            )

        self.at_start(**kwds)
        try:
            Parser(lexer, self).exec()
        finally:
            ret = self.at_end(**kwds)
        return ret

    def parse_file(self, path: str, **kwds: Any) -> Any:
        return self.parse(
            FileLexer(
                path, encoding=self.__class__.encoding,
                command_threshold=self.__class__.__command_threshold__
            ),
            **kwds
        )

    def parse_command(self, cmd: str, **kwds: Any) -> Any:
        return self.parse(
            StringLexer(cmd, stat=1, encoding=self.__class__.encoding), **kwds)

    def parse_args(self, args: str) -> Tuple[tuple, Dict[str, Any]]:
        return Parser(
            StringLexer(args, stat=2, encoding=self.__class__.encoding), self
        ).parse_args()
    
    def __get(self, key: str, default: Optional[Callable]) -> Optional[Callable]:
        # sourcery skip: use-named-expression
        stack = self._stack
        while stack:
            command = stack.command_set.get(key, None)
            if command:
                return command
            stack = stack.next
        return default

    @property
    def top(self) -> Tuple[str, Union[Dict[str, Callable], "KoiLang"]]:
        return self._stack[:2]
    
    def __getitem__(self, key: str) -> Callable:
        # sourcery skip: use-named-expression
        command = self.__get(key, None)
        if not command:
            raise KeyError(f"command '{key}' not found")
        return command
        
    def __contains__(self, key: str) -> bool:
        return self.__get(key, None) is not None

    def __len__(self) -> int:
        length = 0
        stack = self._stack
        while stack:
            length += 1
            stack = stack.next
        return length

    def __repr__(self) -> str:
        return f"<kola {self.__class__.__qualname__} object in environment {self.top[0]}>"


@overload
def kola_command(func: Callable[..., Any], **kwds) -> Command: ...
@overload  # noqa: E302
def kola_command(
    func: Optional[str] = None,
    *,
    method: Literal["static", "class", "default"] = ...,
    envs: Union[Iterable[str], str] = ...,
    alias: Union[Iterable[str], str] = ...
) -> Callable[[Callable[..., Any]], Command]: ...
def kola_command(  # noqa: E302
    func: Union[Callable[..., Any], str, None] = None,
    **kwds
) -> Union[Command, Callable[[Callable], Command]]:
    def wrapper(wrapped_func: Callable[..., Any]) -> Command:
        if isinstance(func, str):
            name = func
        else:
            name = wrapped_func.__name__
        return Command(name, wrapped_func, **kwds)
    if callable(func):
        return wrapper(func)
    else:
        return wrapper


@overload
def kola_text(func: Callable[..., Any], **kwds) -> Command: ...
@overload  # noqa: E302
def kola_text(
    *,
    method: Literal["static", "class", "default"] = ...,
    envs: Union[Iterable[str], str] = ...,
    alias: Union[Iterable[str], str] = ...
) -> Callable[[Callable[..., Any]], Command]: ...
def kola_text(  # noqa: E302
    func: Optional[Callable[..., Any]] = None,
    **kwds
) -> Union[Command, Callable[[Callable], Command]]:
    def wrapper(wrapped_func: Callable[..., Any]) -> Command:
        return Command("@text", wrapped_func, **kwds)
    if callable(func):
        return wrapper(func)
    else:
        return wrapper


@overload
def kola_number(func: Callable[..., Any], **kwds) -> Command: ...
@overload  # noqa: E302
def kola_number(
    *,
    method: Literal["static", "class", "default"] = ...,
    envs: Union[Iterable[str], str] = ...,
    alias: Union[Iterable[str], str] = ...
) -> Callable[[Callable[..., Any]], Command]: ...
def kola_number(  # noqa: E302
    func: Optional[Callable[..., Any]] = None,
    **kwds
) -> Union[Command, Callable[[Callable], Command]]:
    def wrapper(wrapped_func: Callable[..., Any]) -> Command:
        return Command("@number", wrapped_func, **kwds)
    if callable(func):
        return wrapper(func)
    else:
        return wrapper


@overload
def kola_annotation(func: Callable[..., Any], **kwds) -> Command: ...
@overload  # noqa: E302
def kola_annotation(
    *,
    method: Literal["static", "class", "default"] = ...,
    envs: Union[Iterable[str], str] = ...,
    alias: Union[Iterable[str], str] = ...
) -> Callable[[Callable[..., Any]], Command]: ...
def kola_annotation(  # noqa: E302
    func: Optional[Callable[..., Any]] = None,
    **kwds
) -> Union[Command, Callable[[Callable], Command]]:
    def wrapper(wrapped_func: Callable[..., Any]) -> Command:
        return Command("@annotation", wrapped_func, **kwds)
    if callable(func):
        return wrapper(func)
    else:
        return wrapper


@overload
def kola_env(func: Callable[..., Any], **kwds) -> EnvEnter: ...
@overload  # noqa: E302
def kola_env(
    func: Optional[str] = None,
    *,
    env_name: str = ...,
    method: Literal["static", "class", "default"] = ...,
    alias: Union[Iterable[str], str] = ...,
    envs: Union[Iterable[str], str] = ...,
    command_set: Union[Dict[str, Callable], "KoiLang"] = ...
) -> Callable[[Callable[..., Any]], EnvEnter]: ...
def kola_env(  # noqa: E302
    func: Union[Callable[..., Any], str, None] = None,
    *,
    env_name: Optional[str] = None,
    **kwds
) -> Union[EnvEnter, Callable[[Callable], EnvEnter]]:
    def wrapper(wrapped_func: Callable[..., Any]) -> EnvEnter:
        if isinstance(func, str):
            name = func
        else:
            name = wrapped_func.__name__
        return EnvEnter(name, wrapped_func, env_name=env_name, **kwds)
    if callable(func):
        return wrapper(func)
    else:
        return wrapper


@overload
def kola_env_class(kola_cls: Type[KoiLang], **kwds) -> EnvClsEnter: ...
@overload  # noqa: E302
def kola_env_class(
    kola_cls: Optional[str] = None,
    *,
    enter: str = ...,
    alias: Union[Iterable[str], str] = ...,
    exit: Union[str, Iterable[str], None] = ...,
    envs: Union[str, Tuple[str, ...]] = ...
) -> Callable[[Type[KoiLang]], EnvClsEnter]: ...
def kola_env_class(  # noqa: E302
    kola_cls: Union[Type[KoiLang], str, None] = None,
    *,
    enter: str = "enter",
    exit: Union[str, Iterable[str], None] = None,
    **kwds
) -> Union[Callable[[Type[KoiLang]], EnvClsEnter], EnvClsEnter]:
    def wrapper(wrapped_class: Type[KoiLang]) -> EnvClsEnter:
        if isinstance(kola_cls, str):
            env_name = kola_cls
        else:
            env_name = wrapped_class.__name__
        
        enter_func: Callable = getattr(wrapped_class, enter)
        if isinstance(enter_func, Command):
            enter_func.suppression = True
        enter_cmd = EnvClsEnter(
            enter_func.__name__,
            enter_func,
            env_name=env_name,
            command_set_class=wrapped_class,
            exit_entry=exit or (),
            **kwds
        )
        return enter_cmd
    if isinstance(kola_cls, type):
        return wrapper(kola_cls)
    else:
        return wrapper



"""
Copyright 2022 Ovizro

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import os
import re
import sys
import warnings
from setuptools import setup, Extension

try:
    with open("README.md", encoding='utf-8') as f:
        description = f.read()
except OSError:
    warnings.warn("Miss file 'README.md', using default description.", ResourceWarning)
    description = "simple python module for KoiLang parsing"

try:
    with open("kola/version.py") as f:
        version = re.search(r"__version__\s*=\s*\"(.*)\"\n", f.read()).group(1) # type: ignore
except Exception as e:
    raise ValueError("fail to read kola version") from e


USE_CYTHON = "USE_CYTHON" in os.environ
FILE_SUFFIX = ".pyx" if USE_CYTHON else ".c"

extensions = [
    Extension("kola.lexer", ["kola/lexer" + FILE_SUFFIX, "kola/lex.yy.c", "kola/unicode_handler.c"]),
    Extension("kola.parser", ["kola/parser" + FILE_SUFFIX]),
    Extension("kola.writer", ["kola/writer" + FILE_SUFFIX])
]
if USE_CYTHON:
    from Cython.Build import cythonize
    from Cython.Compiler import Options

    Options.fast_fail = True
    extensions = cythonize(extensions, annotate=True, compiler_directives={"language_level": "3"})


setup(
    name="KoiLang",
    version=version,
    description="simple python module for KoiLang parsing",
    long_description=description,
    long_description_content_type='text/markdown',

    author="Ovizro",
    author_email="Ovizro@hypercol.com",
    maintainer="Ovizro",
    maintainer_email="Ovizro@hypercol.com",
    license="Apache 2.0",

    url="https://github.com/Ovizro/Kola",
    packages=["kola"],
    python_requires=">=3.6",
    package_data={'':["*.pyi", "*.pxd", "*.h"]},
    install_requires=[
        "typing_extensions>=4.0" if sys.version_info >= (3, 7)
            else "typing_extensions>=4.0,<4.2"
    ],
    ext_modules=extensions,

    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: C",
        "Programming Language :: Cython",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Text Processing :: Markup"
    ]
)


from .lexer import BaseLexer, FileLexer, StringLexer
from .parser import Parser
from .writer import BaseWriter, FileWriter, StringWriter, BaseWriterItem, FormatItem, ComplexArg, WriterItemLike
from .klvm import KoiLang, kola_command, kola_text, kola_number, kola_env, kola_env_class
from .version import __version__, version_info  # noqa: F401
from .exception import KoiLangError, KoiLangSyntaxError, KoiLangCommandError


__author__ = "Ovizro"
__author_email__ = "Ovizro@hypercol.com"

__all__ = [
    "KoiLang",
    "kola_command",
    "kola_text",
    "kola_number",
    "kola_env",
    "kola_env_class",
    
    "BaseLexer",
    "FileLexer",
    "StringLexer",
    "Parser",
    "BaseWriter",
    "FileWriter",
    "StringWriter",
    "BaseWriterItem",
    "FormatItem",
    "ComplexArg",
    "WriterItemLike",

    "KoiLangError",
    "KoiLangSyntaxError",
    "KoiLangCommandError"
]
